// Directmode Metaballs
// Copyright 2019 SkyBerron
//
// Licensed under the Apache License, Version 2.0 (the "License");
// you may not use this file except in compliance with the License.
// You may obtain a copy of the License at
//
//     http://www.apache.org/licenses/LICENSE-2.0
//
// Unless required by applicable law or agreed to in writing, software
// distributed under the License is distributed on an "AS IS" BASIS,
// WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
// See the License for the specific language governing permissions and
// limitations under the License.


//import femto.mode.HiRes16Color;
import femto.mode.Direct;
import femto.Game;
import femto.State;
import femto.input.Button;
import femto.font.TIC80;
//import static femto.hardware.ST7775.*;
import static femto.hardware.ST7775;
import metaball;

class MBall {
    float xc, yc;
    float vx, vy;

    int x, y;
    int ilut;
}


class Main extends State {
    static final var screen = new Direct(TIC80.font()); // the screenmode we want to draw with
    
    int counter; // variables are automatically initialized to 0 or null
    long timeLast = 0;
    
    int scene = 0;
    
    // Scanline
    short[] scanline;
    
    // Metaball palettes
    short[] m_pal20;
    short[] m_pal30;
    short[] m_pal21;
    short[] m_pal31;

    // Metaballs
    int m_nmb;
    MBall[] m_tmb;

    public static void main(String[] args){
        Game.run( TIC80.font(), new Main() );
    }
    
    void init(){
        screen.clear(0);
        
        // Scanline
        scanline = new short[220];
        
        // Metaball palettes
        m_pal20 = new short[512];
        m_pal30 = new short[384];
        m_pal21 = new short[512];
        m_pal31 = new short[384];

        int[] cmap1 = { 255,255,255, 0, 147,141,224, 18, 0,0,0, 33, 240,47,23, 68, 217,255,0, 100 };
        int[] cmap2 = { 255,255,255, 0, 147,141,224, 18, 245,12,12, 36, 240,47,23, 54, 217,255,0, 100 };
        colormap( cmap1, 5, m_pal20, 512 );
        colormap( cmap1, 5, m_pal30, 384 );
        colormap( cmap2, 5, m_pal21, 512 );
        colormap( cmap2, 5, m_pal31, 384 );

        // Metaballs
        m_nmb = 4;
        m_tmb = new MBall[m_nmb];
        float r0 = 40.0f;

        for( int i = 0; i < m_nmb; i++ ){
            MBall mb = new MBall();
            m_tmb[i] = mb;
            float a0 = ( 2.0f * Math.PI * i ) / m_nmb;
            mb.xc = 110.0f + r0 * Math.cos( a0 );
            mb.yc = 88.0f + r0 * Math.sin( a0 );
            float a1 = 2.0f * Math.PI * Math.random();
            mb.vx = 3.0f * Math.cos( a1 );
            mb.vy = 3.0f * Math.sin( a1 );
        }
        
    }

    void update(){
        int halfw = screen.width()/2;
        int halfh = screen.height()/2;
        String txt = "noname";
        
        if( Button.A.justPressed() ){
            screen.clear(0);
            counter = 0;
        }
        if( Button.B.justPressed() ){
            screen.clear(0);
            scene = ( scene + 1 ) % 6;
        }
        
        if( scene < 6 ) {
            int n = ( scene % 3 ) + 2;
            int m = scene / 3;
            txt = "MetaBalls " + n + " points";

            for( int i = 0; i < m_nmb; i++ ){
                MBall mb = m_tmb[i];
                mb.xc += mb.vx;
                mb.yc += mb.vy;
                if( mb.xc < 0.0f ) {
                    mb.xc = -mb.xc;
                    mb.vx = Math.abs( mb.vx );
                }
                if( mb.yc < 0.0f ) {
                    mb.yc = -mb.yc;
                    mb.vy = Math.abs( mb.vy );
                }
                if( mb.xc >= 220.0f ) {
                    mb.xc = 440.0f - mb.xc;
                    mb.vx = -Math.abs( mb.vx );
                }
                if( mb.yc >= 176.0f ) {
                    mb.yc = 352.0f - mb.yc;
                    mb.vy = -Math.abs( mb.vy );
                }
                
                mb.x = (int) ( .5f + mb.xc );
                mb.y = (int) ( .5f + mb.yc );
            }
            
            pointer ptr = (pointer) metaball.bin();

            short[] pal2 = m_pal20;
            short[] pal3 = m_pal30;
            if( m > 0 ) {
                pal2 = m_pal21;
                pal3 = m_pal31;
            }

            for( int y = 0; y < 176; y++ ) {
                /*
                for( int x = 0; x < 220; x++ ) {
                    scanline[x] = 0;
                }
                */
                
                for( int i = 0; i < m_nmb; i++ ) {
                    MBall mb = m_tmb[i];
                    int dx = 255 - mb.x;
                    int dy = mb.y - y;
                    if( dy >= 0 ) {
                        if( dy > 255 ) dy = 255;
                        mb.ilut = 512 * dy + dx;
                    } else {
                        if( dy < -255 ) dy = -255;
                        mb.ilut = -512 * dy + dx;
                    }
                }

                MBall mb0 = m_tmb[0];
                int ilut0 = mb0.ilut;

                MBall mb1 = m_tmb[1];
                int ilut1 = mb1.ilut;

                MBall mb2 = m_tmb[2];
                int ilut2 = mb2.ilut;

                MBall mb3 = m_tmb[3];
                int ilut3 = mb3.ilut;

                pointer ptr0 = (pointer) metaball.bin() + ilut0;
                pointer ptr1 = (pointer) metaball.bin() + ilut1;
                pointer ptr2 = (pointer) metaball.bin() + ilut2;
                pointer ptr3 = (pointer) metaball.bin() + ilut3;
                
                if( n == 2 ) {
                    for( int x = 0; x < 220; x++ ) {
                        scanline[x] = pal2[ ( (int) System.memory.LDRB( ptr0 + x ) )
                            + ( (int) System.memory.LDRB( ptr1 + x ) )
                        ];
                    }
                } else if( n == 3 ) {
                    for( int x = 0; x < 220; x++ ) {
                        scanline[x] = pal3[ ( ( (int) System.memory.LDRB( ptr0 + x ) )
                            + ( (int) System.memory.LDRB( ptr1 + x ) )
                            + ( (int) System.memory.LDRB( ptr2 + x ) ) ) >> 1
                        ];
                    }
                } else {
                    for( int x = 0; x < 220; x++ ) {
                        scanline[x] = pal2[ ( ( (int) System.memory.LDRB( ptr0 + x ) )
                            + ( (int) System.memory.LDRB( ptr1 + x ) )
                            + ( (int) System.memory.LDRB( ptr2 + x ) )
                            + ( (int) System.memory.LDRB( ptr3 + x ) ) ) >> 1
                        ];
                    }
                    
                }
                
                drawScanline( 0, y, 220, scanline, 0, 220 );
            }
        }

        if( true ) {
            // Print some text
            int y = screen.height() - 6;
            if( scene < 0 ) {
                screen.fillRect( 0, y, screen.width(), 6, 0 );
            }
            screen.setTextPosition( 0, y );
            screen.textColor++;
            screen.println( "Scene " + scene + ": " + txt );
            screen.setTextPosition( screen.width() - 32, y );
            screen.println( (System.currentTimeMillis() - timeLast) + " ms " );
        }

        counter++;
        timeLast = System.currentTimeMillis();
 
    }
    
    // no built-in abs() for floats
    public float abs( float a )   {
        return( ( a >= 0.0f ) ? a : (-a) );
    }
    
    // no built-in % operator for floats
    public float mod( float a, float b ) {
        // using int, casting to long not working
        int d = (int) ( a / b );
        return( a - b * d );
    }
    
    public float mid( float a, float b, float c ) {
        float d = a;
        if( a < b ) {
            if( b < c ) {
                d = b;
            } else if( a < c ) {
                d = c;
            }
        } else if( c < a ) {
            if( b < c ) {
                d = c;
            } else {
                d = b;
            }
        }
        return d;
    }
    
    public int mid( int a, int b, int c ) {
        int d;
        if( a > b ) {
            d = a; a = b; b = d;
        }
        if( b > c ) {
            d = b; b = c; c = d;
        }
        if( a > b ) {
            return a;
        }
        return b;
    }
    
    // no built-in hsl color space conversion
    public int hsl( float h, float s, float l )
    {
        h = mod( h, 360.0f );
        float c = ( 1.0f - abs( 2.0f*l - 1.0f ) ) * s;
        float x = c * ( 1.0f - abs( mod ( h / 60.0f, 2.0f ) - 1.0f ) );
        float m = l - c / 2.0f;
        
        int u = (int) ( .5f + ( c + m ) * 255.0f );
        int v = (int) ( .5f + ( x + m ) * 255.0f );
        
        int r = 0, g = 0, b = 0;
        
        if( h < 180.0f ) {
            if( h < 60.0f ) {
                r = u; g = v;
            } else if( h < 120.0f ) {
                r = v; g = u;
            } else {
                g = u; b = v;
            }
        } else {
            if( h < 240.0f ) {
                g = v; b = u;
            } else if( h < 300.0f ) {
                r = v; b = u;
            } else {
                r = u; b = v;
            }
        }
        
        return (((int)(r>>3))<<11)
                | (((int)(g>>2))<<5)
                | ((int)(b>>3));
    }

    public static void drawScanline( int x, int y, int w, short[] s, int xs, int xf ) {
        if( xs < 0 ) {
            x -= xs;
            w += xs;
            xs = 0;
        }

        if( x < 0 ) {
            xs -= x;
            w += x;
            x = 0;
        }
        
        if( xs + w >= xf ) {
            w = xf - xs;
        }
        
        if( x + w >= 220 )
            w = 220 - x;
        
        if( y < 0 || y >= 176 || x >= 220 || w <= 0 )
            return;

        ST7775.setX(x);
        ST7775.setY(y);
        ST7775.beginStream();
/* */
        int tmp = 0;
        int WRBit = 1<<12;
        int CLR = 252;
        pointer LCD = ST7775.MPIN;
        pointer buffer;
        __inline_cpp__("buffer = s->elements");

        __inline_asm__("
        @output tmp:+l
        @output buffer:+l
        @output w:+l
        @input LCD:l
        @input WRBit:l
        @input CLR:l
        @clobber cc
        
        ldrh @tmp, [@buffer]
        loop%=:
            adds @buffer, 2
            lsls @tmp, 3
            str  @tmp, [@LCD]
            str  @WRBit, [@LCD, @CLR]
            ldrh @tmp, [@buffer]
            subs @w, 1
            str  @WRBit, [@LCD, 124]
        bne loop%=
        ");
/*/
        while( w-- > 0 ) {
            ST7775.writeData( s[xs++] );
        }
        ST7775.setX(0);
        ST7775.setY(0);
/* */
    }


    
    public void dorun( int[] s, int w, int x0, int x1, int v ) {
        if( x0 < 0 )
            x0 = 0;
        if( x1 >= w )
            x1 = w - 1;
        for( int i = x0; i <= x1; i++ ) {
            s[i] = v;
        }
    }
    
    public void colormap( int[] map, int n, short[] pal, int c ) {
        float r0 = 0.0f, g0 = 0.0f, b0 = 0.0f, k0 = 0.0f;
        float k = 0.0f;
        float dk = 100.0f / c;
        
        int i = 0;
        for( int j = 0; j < n; j++ ) {
            int p = j<<2;
            float r1 = (float) map[p];
            float g1 = (float) map[p+1];
            float b1 = (float) map[p+2];
            float k1 = (float) map[p+3];
            for( ; i < c && k < k1; i++ ) {
                float f = ( k - k0 ) / ( k1 - k0 );
                pal[i] = screen.rgb( (int) ( .5f + r0 + f * ( r1 - r0 ) ), (int) ( .5f + g0 + f * ( g1 - g0 ) ), (int) ( .5f + b0 + f * ( b1 - b0 ) ) );
                k += dk;
            }
            r0 = r1;
            g0 = g1;
            b0 = b1;
            k0 = k1;
        }
        for( ; i < c; i++ ) {
            pal[i] = screen.rgb( (int) r0, (int) g0, (int) b0 );
        }
    }

}
